import numpy as np
import copy

# Factorisation A = LU
#  - Entree : A (matrice)
#  - Sortie : L,U (matrices)	
def lu(A):
    (n,p) = A.shape
    if (n != p) :
        print "Probleme dimension"
        return
    
    L = np.eye(n)
    U = np.zeros((n,p))
    for k in range(n):
        for c in range(k,n):
            U[k,c] = A[k,c]
            for j in range(k):
                U[k,c] = U[k,c] - L[k,j]*U[j,c]
        for l in range(k+1,n):
            L[l,k] = A[l,k]
            for j in range(k):
                L[l,k] = L[l,k] - L[l,j]*U[j,k]
            L[l,k] = L[l,k] *1.0/ U[k,k]
    
    return [L,U]
    
    
    
# Factorisation cholesky A = GG^t
#  - Entree : A (matrice)
#  - Sortie : G (matrice)	    
def cholesky(A):
    (n,p) = A.shape
    if (n != p) :
        print "Probleme dimension matrice non carree"
        return
    
    
    G = np.zeros((n,p))
    for k in range(n):
        # Completer ICI ;
        #G[k,k] = ...
        #...
    
    return G
    
    
# Resolution de U x = b
#  - Entree : U (matrice), b (vecteur)
#  - Sortie : x (Vecteur)
def remontee(U,b):
    (n,p) = U.shape
    x = np.zeros(n)
    for l in range(n-1,-1,-1):
        x[l] = b[l]
        for c in range(l+1,n):
            x[l] = x[l] - U[l,c]*x[c]
        if (np.abs(U[l,l]) > 1e-8):
            x[l] = x[l] / U[l,l]
        else:
            x[l] = x[l]
    return x
    
    
# Resolution de L x = b
#  - Entree : L (matrice), b (vecteur)
#  - Sortie : x (Vecteur)    
def descente(L,b):
    (n,p) = L.shape
    x = np.zeros(n)
    for l in range(n):
        x[l] = b[l]
        for c in range(l):
            x[l] = x[l] - L[l,c]*x[c]
        x[l] = x[l] / L[l,l]
    
    return x   


# Resolution de A x = b par descente de gradient
#  - Entree : A (matrice), b (vecteur), x0 (vecteur), eps (double), iteMax (int), n (int), affich (int)
#  - Sortie : x (Vecteur), residus (tableau double)    
def gradient(A,b,x0,eps,iteMax,n,affich):    
    (nn,p) = A.shape
    
    # Initialisation :
    r = np.dot(A,x0) - b
    residus = [np.dot(r,r)]
    ite = 1
    x = np.copy(x0)
    
    while(residus[-1] >= eps*eps*residus[0] and iteMax > ite):
        if(affich == 1):
            print "Ite : ",ite," / ",iteMax 
            print "Norme residu : ",residus[-1]/residus[0]
        # Completer ICI :
        #Ar = ...
        #...
        
                
        residus.append(np.dot(r,r))
        ite += 1
        
    print "Ite : ",ite," / ",iteMax 
    print "Norme residu : ",residus[-1]/residus[0]
    
    return [x,residus] 

# Resolution de A x = b par gradient conjugue
#  - Entree : A (matrice), b (vecteur), x0 (vecteur), eps (double), iteMax (int), n (int), affich (int)
#  - Sortie : x (Vecteur), residus (tableau double)  
def CG(A,b,x0,eps,iteMax,n,affich):    
    (nn,p) = A.shape
    
    # Initialisation :
    r = np.dot(A,x0) - b
    w = np.copy(r)
    residus = [np.dot(r,r)]
    ite = 1
    x = np.copy(x0)
    
    while(residus[-1] >= eps*eps*residus[0] and iteMax > ite):
        if(affich == 1):
            print "Ite : ",ite," / ",iteMax 
            print "Norme residu : ",residus[-1]/residus[0]
        # Completer ICI :
        Aw = np.dot(A,w)
        #z = ...
        #...
        
        residus.append(np.dot(r,r))
        ite += 1
        
    print "Ite : ",ite," / ",iteMax 
    print "Norme residu : ",residus[-1]/residus[0]
    
    return [x,residus] 



# Resolution de A x = b par la methode de votre choix :
# mise sous la forme de Hessenberg : A COMPLETER !!
def mamethode(A,b):    
    (n,p) = A.shape
    x = b*0.0
    
    # Completer ICI :

    return x


# Decomposition QR
#  - Entree : A (matrice nxp)
#  - Sortie : Q (matrice nxp), R (matrice pxp)
def qr(A):
    (n,p) = A.shape
    # Declaration des variables et initialisation
    Q = np.zeros((n,p))
    R = np.zeros((p,p))
    
    
    # Orthonormalisation G.S.
    for i in range(p):
        v = A[:,i]
        for j in range(i):
            #Completer ICI :
            #scal = ...
            #v = ...
            R[j,i] = scal
        
        normV = np.sqrt(np.dot(v,v))
        # Completer ICI :
        #R[i,i] = ...
        
        # On verifie que le vecteur v est de norme > 0
        if (normV > 1e-8):
            Q[:,i] = v / normV
            
    return [Q,R]
        
    
# Reconstruction source u:
#  - Entree : coeff (vecteur), x (vecteur), N (entier)
#  - Sortie : y (vecteur) 
def reconsU(coeff,x,N):
    y = x*0
    
    for i in range(N+1):
        c = coeff[i]
        y = y + np.cos(2*np.pi*x*(i-1))*c
    for i in range(N):
        s = coeff[i+N+1]
        y = y + np.sin(2*np.pi*x*i)*s
    
    return y
    
    
# Reconstruction source f:
#  - Entree : coeff (vecteur), x (vecteur), N (entier), omega (reel)
#  - Sortie : y (vecteur) 
def reconsF(coeff,x,N,omega):
    y = x*0
    
    for i in range(N+1):
        c = coeff[i]
        # Completer ICI :
        #y = y - ...
    for i in range(N):
        s = coeff[i+N+1]
        # Completer ICI :
        #y = y - ...
    
    return y